-------------------------------------------------------------------------------
-- Finders Keepers By Lockslap (US, Bleeding Hollow)
--------------------------------------------------------------------------------
local FK = LibStub("AceAddon-3.0"):NewAddon("Finders Keepers", "AceEvent-3.0", "AceConsole-3.0")
if not TomTom then return end

-- libraries
local L = LibStub("AceLocale-3.0"):GetLocale("Finders Keepers", false)
local Astrolabe = DongleStub("Astrolabe-1.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

-- globals
FK.name = GetAddOnMetadata("FindersKeepers", "Title")
FK.version = GetAddOnMetadata("FindersKeepers", "Version")
FK.waypoints = {}

local zones = {
	[806] = true,  -- The Jade Forest
	[807] = true,  -- Valley of the Four Winds
	[809] = true,  -- Kun-Lai Summit
	[810] = true,  -- Townlong Steppes
	[811] = true,  -- Vale of Eternal Blossoms
	[857] = true,  -- Krasarang Wilds
	[858] = true,  -- Dread Wastes
	[873] = true,  -- The Veiled Stair
}

-- options
local db, defaults = {}, {
	profile = {
		auto = false,
		notify = false,
		completed = false,
		filters = {
			objects = true,
			npcs = true,
			containers = true,
		},
		tomtom = {
			minimap = true,
			worldmap = true,
			arrow = true,
		},
		npcscan = {
			["31307"] = false,
			["31432"] = false,
			["31284"] = false,
			["31292"] = false,
		},
	
		-- for the initial loading
		firstLoad = false,
		
		-- waypoints
		treasures = {
			objects = {},
			npcs = {},
			containers = {},
		},
	}
}

local fields = {"Title", "Version", "Author", "X-Category", "X-License", "X-Email", "Email", "eMail", "X-Website", "X-Credits", "X-Localizations", "X-Donate", "X-Paypal"}
local fieldName = "|cff9382c9%s:|r"
local options = {
	type = "group",
	args = {
		general = {
			type = "group",
			name = FK.name,
			get = function(info) return db[info[#info]] end,
			set = function(info, value)
				db[info[#info]] = value
				FK:RemoveWaypoints()
			end,
			args = {
				header = {
					type = "header",
					order = 1,
					name = (L["Version: |cff99ff33%s|r"]):format(FK.version),
					width = "full",
				},
				description = {
					type = "description",
					order = 2,
					name = L["A simple addon to help you find the various treasures hidden throughout Pandaria."],
					width = "full",
				},
				notify = {
					type = "toggle",
					order = 3,
					name = L["Print Notices"],
					desc = L["Send various notifications when the addon completes an action."],
				},
				completed = {
					type = "toggle",
					order = 4,
					name = L["Completed Waypoints"],
					desc = L["Include completed waypoints when determining the waypoints to be added to TomTom.\n\nThis applies to automatic and manual mode."],
				},
				tomtom = {
					type = "group",
					order = 100,
					name = L["Add Waypoint To..."],
					guiInline = true,
					get = function(info) return db.tomtom[info[#info]] end,
					set = function(info, value)
						db.tomtom[info[#info]] = value
						FK:RemoveWaypoints()
					end,
					args = {
						arrow = {
							type = "toggle",
							order = 1,
							name = L["Crazy Arrow"],
							desc = L["Show waypoint on TomTom's crazy arrow."],
						},
						minimap = {
							type = "toggle",
							order = 2,
							name = L["Minimap"],
							desc = L["Show waypoint on the minimap."],
						},
						worldmap = {
							type = "toggle",
							order = 3,
							name = L["World Map"],
							desc = L["Show waypoint on the world map."],
						},	
					},
				},
				filters = {
					type = "group",
					order = 101,
					name = L["Waypoint Filters"],
					guiInline = true,
					get = function(info) return db.filters[info[#info]] end,
					set = function(info, value)
						db.filters[info[#info]] = value
						FK:RemoveWaypoints()
					end,
					args = {
						objects = {
							type = "toggle",
							order = 1,
							name = L["Object (Items)"],
							desc = L["Include objects (items) in the list of waypoints loaded into TomTom."],
						},
						npcs = {
							type = "toggle",
							order = 2,
							name = L["NPCs"],
							desc = L["Include NPCs in the list of waypoints loaded into TomTom."],
						},
						containers = {
							type = "toggle",
							order = 3,
							name = L["Containers"],
							desc = L["Include containers in the list of waypoints loaded into TomTom."],
						},
					},
				},
			},
		},
		npcscan = {
			type = "group",
			name = L["NPC Scan"],
			disabled = function() return not IsAddOnLoaded("_NPCScan") end,
			get = function(info) return db.npcscan[info[#info]] end,
			set = function(info, value)
				db.npcscan[info[#info]] = value
				local id = tonumber(info[#info])
				local treasure = db.treasures.npcs[id] or nil
				if not _NPCScan or not treasure then return end
				if db.npcscan[info[#info]] then
					_NPCScan.NPCAdd(treasure.npcid, treasure.npc)
					if db.notify then
						FK:Print((L["Added |cff00ff96%s|r to NPC Scan."]):format(treasure.npc))
					end
				else
					_NPCScan.NPCRemove(treasure.npcid)
					if db.notify then
						FK:Print((L["Removed |cff00ff96%s|r from NPC Scan."]):format(treasure.npc))
					end
				end
			end,
			args = {
				about = {
					type = "description",
					order = 1,
					name = L["Here you can choose to have the addon automatically add the NPCs required for 4 of the treasures to NPC Scan."],
					width = "full",
				},
				["31307"] = {
					type = "toggle",
					name = L["Jade Warrior Statue"],
					desc = L["Add/remove the Jade Warrior Statue to NPC Scan's list of NPCs to look for."],
					width = "full",
					disabled = function() return db.treasures.npcs[31307].completed end,
				},
				["31432"] = {
					type = "toggle",
					name = L["Glinting Rapana Whelk"],
					desc = L["Add/remove the Glinting Rapana Whelk to NPC Scan's list of NPCs to look for."],
					width = "full",
					disabled = function() return db.treasures.npcs[31432].completed end,
				},
				["31284"] = {
					type = "toggle",
					name = L["Ghostly Padaren Fisherman"],
					desc = L["Add/remove the Ghostly Pandaren Fisherman to NPC Scan's list of NPCs to look for."],
					width = "full",
					disabled = function() return db.treasures.npcs[31284].completed end,
				},
				["31292"] = {
					type = "toggle",
					name = L["Ghostly Pandaren Craftsman"],
					desc = L["Add/remove the Ghostly Pandaren Craftsman to NPC Scan's list of NPCs to look for."],
					width = "full",
					disabled = function() return db.treasures.npcs[31292].completed end,
				},
			},
		},
		about = {
			type = "group",
			name = L["About"],
			order = -1,
			args = {},
		},
	},
}

function FK:InArray(value)
	if #self.waypoints == 0 then return false end
	for i = 1, #self.waypoints do
		if self.waypoints[i] == value then
			return i
		end
	end
	return false
end

function FK:GetPossibleWaypoints()
	local waypoints, count = {}, 0
	local map, _, _, _ = Astrolabe:GetCurrentPlayerPosition()

	if db.treasures.npcs then
		for index, value in pairs(db.treasures.npcs) do
			if map == value.zone then
				if db.completed or (not db.completed and not value.completed) then
					waypoints[index] = value
					count = count + 1
				end
			end
		end
	end
	if db.treasures.objects then
		for index, value in pairs(db.treasures.objects) do
			if map == value.zone then
				if db.completed or (not db.completed and not value.completed) then
					waypoints[index] = value
					count = count + 1
				end
			end
		end
	end
	if db.treasures.containers then
		for index, value in pairs(db.treasures.containers) do
			if map == value.zone then
				if db.completed or (not db.completed and not value.completed) then
					waypoints[index] = value
					count = count + 1
				end
			end
		end
	end
	
	return count > 0 and waypoints or false
end

function FK:RemoveWaypoints()
	if not self.waypoints or #self.waypoints == 0 then return end
	for i = 1, #self.waypoints do
		if self.waypoints[i] then
			TomTom:RemoveWaypoint(self.waypoints[i])
		end
	end
end

function FK:LoadWaypoints()
	local map, _, _, _ = Astrolabe:GetCurrentPlayerPosition()
	if not zones[map] then
		if db.notify then
			self:Print(L["You must be in |cff00ff96Pandaria|r to use waypoints."])
		end
		return
	end
	
	-- pull the possible waypoints based on completions and settings
	local possible = self:GetPossibleWaypoints()
	if not possible then
		if db.notify then
			self:Print(L["No waypoints match your settings."])
		end
		return
	end
	
	local count, opts = 0, {
		title = "",
		persistent = false,
		minimap = db.tomtom.minimap,
		world = db.tomtom.worldmap,
		crazy = db.tomtom.arrow,
		cleardistance = 0,
		arrivaldistance = 0,
		callbacks = {},
	}
	opts.callbacks = TomTom:DefaultCallbacks(opts)
	opts.callbacks.distance[15] = function(event, uid, range, distance, lastDistance)
		if lastDistance > distance then
			local inArray = FK:InArray(uid)
			TomTom:RemoveWaypoint(uid)
			if inArray ~= false then
				self.waypoints[inArray] = nil
			end
			
			if db.notify and #waypoints == 0 then
				self:Print((L["You have completed all the waypoints for |cff00ff96%s|r."]):format(GetZoneText()))
			end
		end
	end
	for index, value in pairs(possible) do
		if #value.coords == 1 then
			opts.title = ("%s\n%s"):format(value.name, value.zoneName)
			self.waypoints[#self.waypoints + 1] = TomTom:AddMFWaypoint(map, 0, value.coords[1].x / 100, value.coords[1].y / 100, opts)
			count = count + 1
		else
			-- for treasures with multiple locations
			for i, coords in ipairs(value.coords) do
				opts.title = ("%s #%d\n%s"):format(value.name, i, value.zoneName)
				self.waypoints[#self.waypoints + 1] = TomTom:AddMFWaypoint(map, 0, coords.x / 100, coords.y / 100, opts)
				count = count + 1
			end
		end
	end
	
	if db.notify then
		self:Print((L["Added |cff00ff96%d|r waypoints for |cff00ff96%s|r."]):format(count, GetZoneText()))
	end
end

function FK:LoadTreasuresIntoDatabase()
	local count = 0
	-- first load the objects (items)
	db.treasures.objects = {}
	for quest, treasure in pairs(FK.objects) do
		-- see if they've completed the quest
		local completed = IsQuestFlaggedCompleted(quest) and true or false
		local name, link, _, _, _, _, _, _, _, texture, _ = GetItemInfo(treasure.item)
		if not db.treasures.objects[quest] then db.treasures.objects[quest] = {} end
		db.treasures.objects[quest] = {
			completed = completed,
			item = treasure.item,
			name = name,
			link = link,
			icon = texture,
			zone = treasure.zone,
			zoneName = GetMapNameByID(treasure.zone),
			coords = treasure.coords,
		}
		count = count + 1
	end	
	if db.notify then
		self:Print((L["Added |cff00ff96%d|r %s into the database."]):format(count, L["objects"]))
	end
	
	-- now npcs
	count = 0
	db.treasures.npcs = {}
	for quest, treasure in pairs(FK.npcs) do
		-- see if they've completed the quest
		local completed = IsQuestFlaggedCompleted(quest) and true or false
		local name, link, _, _, _, _, _, _, _, texture, _ = GetItemInfo(treasure.item)
		if not db.treasures.npcs[quest] then db.treasures.npcs[quest] = {} end
		db.treasures.npcs[quest] = {
			completed = completed,
			item = treasure.item,
			name = name,
			link = link,
			icon = texture,
			zone = treasure.zone,
			zoneName = GetMapNameByID(treasure.zone),
			coords = treasure.coords,
			npc = treasure.npc,
			npcid = treasure.npcid,
		}
		count = count + 1
	end
	if db.notify then
		self:Print((L["Added |cff00ff96%d|r %s into the database."]):format(count, L["NPCs"]))
	end
	
	-- finally containers
	count = 0
	db.treasures.containers = {}
	for quest, treasure in pairs(FK.containers) do
		-- see if they've completed the quest
		local completed = IsQuestFlaggedCompleted(quest) and true or false
		if not db.treasures.containers[quest] then db.treasures.containers[quest] = {} end
		db.treasures.containers[quest] = {
			name = treasure.name,
			completed = completed,
			zone = treasure.zone,
			zoneName = GetMapNameByID(treasure.zone),
			coords = treasure.coords,
		}
		count = count + 1
	end
	if db.notify then
		self:Print((L["Added |cff00ff96%d|r %s into the database."]):format(count, L["containers"]))
	end
end

function FK:UpdateTreasures()
	for index, value in pairs(db.treasures.objects) do
		db.treasures.objects[index].completed = IsQuestFlaggedCompleted(index) and true or false
	end
	
	for index, value in pairs(db.treasures.npcs) do
		db.treasures.npcs[index].completed = IsQuestFlaggedCompleted(index) and true or false
	end
	
	for index, value in pairs(db.treasures.containers) do
		db.treasures.containers[index].completed = IsQuestFlaggedCompleted(index) and true or false
	end
	
	if db.notify then
		self:Print(L["Updated treasure objects, NPCs, and containers."])
	end
end

function FK:HandleChatCommand(args)
	local command, value = self:GetArgs(args, 2)
	
	if command == "config" then
		InterfaceOptionsFrame_OpenToCategory(self.frames.general)
	
	elseif command == "reload" then
		self:LoadTreasuresIntoDatabase()
	
	elseif command == "load" then
		self:LoadWaypoints()
	
	elseif command == "update" then
		self:UpdateTreasures()
	
	elseif command == "clear" then
		self:RemoveWaypoints()
		self:Print(L["Waypoints cleared."])
		
	elseif not command or command == "" or command == "help" or command == "?" then
		self:Print((L["Finders Keepers version |cff00ff96%s|r by |cff9382c9Lockslap|r"]):format(self.version))
		print(L["Chat command:  |cff00ff96/finders|r or |cff00ff96/fk|r"])
		print(L["    |cff00ff96config|r - Open configuration window."])
		print(L["    |cff00ff96help|r or |cff00ff96?|r - Print this help again."])
		print(L["    |cff00ff96reload|r - Clear and reload the treasure database."])
		print(L["    |cff00ff96update|r - Update the treasure database."])
		print(L["    |cff00ff96load|r - Load the waypoints for your current zone."])
		print(L["    |cff00ff96clear|r - Clear the waypoints."])
	end
end

function FK:OnEnable()
	self:RegisterChatCommand("finders", "HandleChatCommand")
	self:RegisterChatCommand("fk", "HandleChatCommand")
end

function FK:OnDisable()
	self:UnregisterChatCommand("finders")
	self:UnregisterChatCommand("fk")
	
	if self.waypoints then
		self:RemoveWaypoints()
	end
	
	if self.waypoint then
		TomTom:RemoveWaypoint(self.waypoint)
		self.waypoint = nil
	end
end

function FK:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("FindersKeepersDB", defaults)
	db = self.db.profile
	
	-- setup tomtom
	TomTom.profile.arrow.enable = true
	TomTom.profile.arrow.arrival = 25
	TomTom.profile.arrow.setclosest = true
	
	-- load/update the waypoints
	if not db.firstLoad then
		self:LoadTreasuresIntoDatabase()
		db.firstLoad = true
	else
		self:UpdateTreasures()
	end
	
	-- about page
	local i = 1
	for _, field in pairs(fields) do
		local value = GetAddOnMetadata(self.name:gsub(" ", ""), field)
		if value then
			options.args.about.args[field] = {
				type = "description",
				order = i,
				width = "normal",
				name = fieldName:format(field:gsub("X%-", "")),
			}
			i = i + 1
			options.args.about.args[("%s_value"):format(field)] = {
				type = "description",
				order = i,
				width = "double",
				name = value
			}
			i = i + 1
		end
	end
	
	-- register the config window with the ui
	AceConfigRegistry:RegisterOptionsTable(self.name, options)
	self.frames = {}
	self.frames.general = AceConfigDialog:AddToBlizOptions(self.name, nil, nil, "general")
	self.frames.npcscan = AceConfigDialog:AddToBlizOptions(self.name, L["NPC Scan"], self.name, "npcscan")
	self.frames.about = AceConfigDialog:AddToBlizOptions(self.name, L["About"], self.name, "about")
end